from rest_framework import status

from mayan.apps.documents.tests.base import GenericDocumentAPIViewTestCase
from mayan.apps.documents.permissions import (
    permission_document_type_edit, permission_document_type_view
)

from ..events import event_index_template_created, event_index_template_edited
from ..models.index_instance_models import IndexInstanceNode, IndexTemplate, IndexTemplateNode
from ..permissions import (
    permission_index_template_create, permission_index_template_delete,
    permission_index_template_edit, permission_index_template_rebuild,
    permission_index_template_view
)

from .literals import TEST_INDEX_TEMPLATE_LABEL
from .mixins import (
    IndexInstanceTestMixin, IndexTemplateTestMixin,
    IndexTemplateActionAPIViewTestMixin, IndexTemplateAPIViewTestMixin,
    IndexTemplateDocumentTypeAPIViewTestMixin, IndexTemplateNodeAPITestMixin
)


class IndexTemplateAPIViewTestCase(
    IndexTemplateTestMixin, IndexTemplateAPIViewTestMixin,
    GenericDocumentAPIViewTestCase
):
    auto_create_test_index_template = False
    auto_upload_test_document = False

    def test_index_template_create_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_index_template_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(IndexTemplate.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_create_api_view_with_permission(self):
        self.grant_permission(permission=permission_index_template_create)

        self._clear_events()

        response = self._request_test_index_template_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.data['id'], self._test_index_template.pk)
        self.assertEqual(
            response.data['label'], self._test_index_template.label
        )

        self.assertEqual(IndexTemplate.objects.count(), 1)
        self.assertEqual(
            self._test_index_template.label, TEST_INDEX_TEMPLATE_LABEL
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_created.id)

    def test_index_template_delete_api_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self._test_index_template in IndexTemplate.objects.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_delete_api_view_with_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_delete
        )

        self._clear_events()

        response = self._request_test_index_template_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertTrue(
            self._test_index_template not in IndexTemplate.objects.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_detail_api_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('id' not in response.data)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_detail_api_view_with_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_view
        )

        self._clear_events()

        response = self._request_test_index_template_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self._test_index_template.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_edit_via_patch_api_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('id' not in response.data)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_edit_via_patch_api_view_with_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self._test_index_template.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)

    def test_index_template_list_api_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_list_api_view_with_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_view
        )

        self._clear_events()

        response = self._request_test_index_template_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'], self._test_index_template.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class IndexTemplateDocumentTypeAPIViewTestCase(
    IndexTemplateTestMixin, IndexTemplateDocumentTypeAPIViewTestMixin,
    GenericDocumentAPIViewTestCase
):
    auto_upload_test_document = False

    def test_index_template_document_type_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_index_template_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_list_api_view_with_document_type_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_view
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_list_api_view_with_index_template_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_view
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_view
        )
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_view
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'], self._test_document_type.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_api_view_no_permission(self):
        self._test_document_type.index_templates.remove(
            self._test_index_template
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self._test_document_type not in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_api_view_with_document_type_access(self):
        self._test_document_type.index_templates.remove(
            self._test_index_template
        )

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self._test_document_type not in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_api_view_with_index_template_access(self):
        self._test_document_type.index_templates.remove(
            self._test_index_template
        )

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertTrue(
            self._test_document_type not in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_api_view_with_full_access(self):
        self._test_document_type.index_templates.remove(
            self._test_index_template
        )

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertTrue(
            self._test_document_type in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)

    def test_index_template_document_type_remove_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_index_template_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self._test_document_type in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_remove_api_view_with_document_type_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self._test_document_type in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_remove_api_view_with_index_template_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertTrue(
            self._test_document_type in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_remove_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertTrue(
            self._test_document_type not in self._test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)


class IndexTemplateRebuildAPIViewTestCase(
    IndexInstanceTestMixin, IndexTemplateTestMixin,
    IndexTemplateActionAPIViewTestMixin, GenericDocumentAPIViewTestCase
):
    auto_create_test_document_stub = True
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        IndexInstanceNode.objects.exclude(parent=None).delete()

    def test_index_template_rebuild_api_view_no_permission(self):
        test_index_instance_node_count = self._test_index_instance.get_descendants().count()

        self._clear_events()

        response = self._request_test_index_template_rebuild_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_index_instance.get_descendants().count(),
            test_index_instance_node_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_rebuild_api_view_with_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_rebuild
        )

        test_index_instance_node_count = self._test_index_instance.get_descendants().count()

        self._clear_events()

        response = self._request_test_index_template_rebuild_api_view()

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            self._test_index_instance.get_descendants().count(),
            test_index_instance_node_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class IndexTemplateResetAPIViewTestCase(
    IndexInstanceTestMixin, IndexTemplateTestMixin,
    IndexTemplateActionAPIViewTestMixin, GenericDocumentAPIViewTestCase
):
    auto_create_test_document_stub = True
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()

        self._test_index_template.rebuild()

    def test_index_template_reset_api_view_no_permission(self):
        test_index_instance_node_count = self._test_index_instance.get_descendants().count()

        self._clear_events()

        response = self._request_test_index_template_reset_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_index_instance.get_descendants().count(),
            test_index_instance_node_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_reset_api_view_with_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_rebuild
        )

        test_index_instance_node_count = self._test_index_instance.get_descendants().count()

        self._clear_events()

        response = self._request_test_index_template_reset_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            self._test_index_instance.get_descendants().count(),
            test_index_instance_node_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class IndexTemplateNodeAPIViewTestCase(
    IndexTemplateTestMixin, IndexTemplateNodeAPITestMixin,
    GenericDocumentAPIViewTestCase
):
    auto_upload_test_document = False

    def test_index_template_node_no_parent_create_api_view_no_permission(self):
        IndexTemplateNode.objects.all().delete()

        index_template_node_count = IndexTemplateNode.objects.count()

        self._clear_events()

        response = self._request_test_index_template_node_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_no_parent_create_api_view_with_access(self):
        IndexTemplateNode.objects.all().delete()

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )
        index_template_node_count = IndexTemplateNode.objects.count()

        self._clear_events()

        response = self._request_test_index_template_node_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_bad_parent_create_api_view_no_permission(self):
        self._create_test_index_template()

        index_template_node_count = IndexTemplateNode.objects.count()

        self._clear_events()

        response = self._request_test_index_template_node_create_api_view(
            extra_data={
                'parent': self._test_index_templates[0].index_template_root_node.pk
            }
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_bad_parent_create_api_view_with_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )
        index_template_node_count = IndexTemplateNode.objects.count()

        self._clear_events()

        response = self._request_test_index_template_node_create_api_view(
            extra_data={
                'parent': self._test_index_templates[0].index_template_root_node.pk
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_create_api_view_no_permission(self):
        index_template_node_count = IndexTemplateNode.objects.count()

        self._clear_events()

        response = self._request_test_index_template_node_create_api_view(
            extra_data={
                'parent': self._test_index_template.index_template_root_node.pk
            }
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_create_api_view_with_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )
        index_template_node_count = IndexTemplateNode.objects.count()

        self._clear_events()

        response = self._request_test_index_template_node_create_api_view(
            extra_data={
                'parent': self._test_index_template.index_template_root_node.pk
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(
            response.data['id'], self._test_index_template_node.pk
        )

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_delete_api_view_no_permission(self):
        index_template_node_count = IndexTemplateNode.objects.count()

        self._clear_events()

        response = self._request_test_index_template_node_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_delete_api_view_with_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )
        index_template_node_count = IndexTemplateNode.objects.count()

        self._clear_events()

        response = self._request_test_index_template_node_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_detail_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_index_template_node_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('count' not in response.data)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_detail_api_view_with_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_view
        )

        self._clear_events()

        response = self._request_test_index_template_node_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self._test_index_template_node.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_edit_via_patch_api_view_no_permission(self):
        index_template_node_expression = self._test_index_template_node.expression

        self._clear_events()

        response = self._request_test_index_template_node_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_index_template_node.refresh_from_db()
        self.assertEqual(
            index_template_node_expression,
            self._test_index_template_node.expression
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_edit_via_patch_api_view_with_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_edit
        )
        index_template_node_expression = self._test_index_template_node.expression

        self._clear_events()

        response = self._request_test_index_template_node_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_index_template_node.refresh_from_db()
        self.assertNotEqual(
            index_template_node_expression,
            self._test_index_template_node.expression
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_index_template_node_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('count' not in response)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_node_list_api_view_with_access(self):
        self.grant_access(
            obj=self._test_index_template,
            permission=permission_index_template_view
        )

        self._clear_events()

        response = self._request_test_index_template_node_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'],
            self._test_index_template_node.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import logging
import sys
import traceback

from django.apps import apps
from django.conf import settings
from django.db import models, transaction

from .events import (
    event_parsing_document_file_content_deleted,
    event_parsing_document_file_finished
)
from .parsers import Parser

logger = logging.getLogger(name=__name__)


class DocumentFilePageContentManager(models.Manager):
    def delete_content_for(self, document_file, user=None):
        with transaction.atomic():
            for document_file_page in document_file.pages.all():
                self.filter(document_file_page=document_file_page).delete()

            event_parsing_document_file_content_deleted.commit(
                actor=user, action_object=document_file.document,
                target=document_file
            )

    def process_document_file(self, document_file, user=None):
        logger.info(
            'Starting parsing for document file: %s', document_file
        )
        logger.debug('document file: %d', document_file.pk)
        try:
            Parser.parse_document_file(document_file=document_file)

            logger.info(
                'Parsing complete for document file: %s', document_file
            )
            document_file.error_log.all().delete()

            event_parsing_document_file_finished.commit(
                action_object=document_file.document, actor=user,
                target=document_file
            )
        except Exception as exception:
            logger.error(
                'Parsing error for document file: %d; %s',
                document_file.pk, exception, exc_info=True
            )

            if settings.DEBUG:
                result = []
                type, value, tb = sys.exc_info()
                result.append('{}: {}'.format(type.__name__, value))
                result.extend(traceback.format_tb(tb))
                document_file.error_log.create(
                    text='\n'.join(result)
                )
            else:
                document_file.error_log.create(text=exception)


class DocumentTypeSettingsManager(models.Manager):
    def get_by_natural_key(self, document_type_natural_key):
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        try:
            document_type = DocumentType.objects.get_by_natural_key(document_type_natural_key)
        except DocumentType.DoesNotExist:
            raise self.model.DoesNotExist

        return self.get(document_type__pk=document_type.pk)



from django.db import models

from mayan.apps.testing.tests.base import BaseTestCase

from ..classes import QuerysetParametersSerializer


class QuerysetParametersSerializerTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.TestModelParent = self._create_test_model(
            model_name='TestModelParent'
        )
        self.TestModelChild = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelParent'
                )
            }, model_name='TestModelChild'
        )

        self._test_object_parent = self.TestModelParent.objects.create()
        self.TestModelChild.objects.create(parent_id=self._test_object_parent.pk)

    def _assertQuerysetEqual(self):
        rebuilt_items = list(map(repr, self.queryset_rebuilt))

        self.assertQuerysetEqual(
            qs=self.queryset_original, values=rebuilt_items
        )

    def test_without_kwargs(self):
        self.queryset_original = self.TestModelParent.objects.all()

        decomposed_queryset = QuerysetParametersSerializer.decompose(
            _model=self.TestModelParent, _method_name='all'
        )

        self.queryset_rebuilt = QuerysetParametersSerializer.rebuild(
            decomposed_queryset=decomposed_queryset
        )

        self._assertQuerysetEqual()

    def test_foreign_key_model(self):
        self.queryset_original = self.TestModelChild.objects.all()

        decomposed_queryset = QuerysetParametersSerializer.decompose(
            _model=self.TestModelChild, _method_name='filter',
            parent=self._test_object_parent
        )

        self.queryset_rebuilt = QuerysetParametersSerializer.rebuild(
            decomposed_queryset=decomposed_queryset
        )

        self._assertQuerysetEqual()

    def test_foreign_key_model_id_query(self):
        self.queryset_original = self.TestModelChild.objects.all()

        decomposed_queryset = QuerysetParametersSerializer.decompose(
            _model=self.TestModelChild, _method_name='filter',
            parent_id=self._test_object_parent.pk
        )

        self.queryset_rebuilt = QuerysetParametersSerializer.rebuild(
            decomposed_queryset=decomposed_queryset
        )

        self._assertQuerysetEqual()



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_('File caching'), name='file_caching'
)

event_cache_created = namespace.add_event_type(
    label=_('Cache created'), name='cache_created'
)
event_cache_edited = namespace.add_event_type(
    label=_('Cache edited'), name='cache_edited'
)
event_cache_purged = namespace.add_event_type(
    label=_('Cache purged'), name='cache_purged'
)
event_cache_partition_purged = namespace.add_event_type(
    label=_('Cache partition purged'), name='cache_partition_purged'
)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='index',
            name='name',
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from ..icons import (
    icon_document_version_page_delete, icon_document_version_page_list,
    icon_document_version_page_list_remap,
    icon_document_version_page_navigation_first,
    icon_document_version_page_navigation_last,
    icon_document_version_page_navigation_next,
    icon_document_version_page_navigation_previous,
    icon_document_version_page_return_to_document,
    icon_document_version_page_return_to_document_version,
    icon_document_version_page_return_to_document_version_page_list,
    icon_document_version_page_rotate_left,
    icon_document_version_page_rotate_right,
    icon_document_version_page_detail, icon_document_version_page_detail_reset,
    icon_document_version_page_zoom_in, icon_document_version_page_zoom_out
)
from ..permissions import (
    permission_document_version_edit, permission_document_version_view,
    permission_document_view
)
from ..settings import setting_zoom_max_level, setting_zoom_min_level


def is_first_page(context):
    return context['resolved_object'].siblings.first() == context['resolved_object']


def is_last_page(context):
    return context['resolved_object'].siblings.last() == context['resolved_object']


def is_max_zoom(context):
    return context['zoom'] >= setting_zoom_max_level.value


def is_min_zoom(context):
    return context['zoom'] <= setting_zoom_min_level.value


link_document_version_page_delete = Link(
    args='resolved_object.pk', icon=icon_document_version_page_delete,
    permissions=(permission_document_version_edit,), tags='dangerous',
    text=_('Delete'), view='documents:document_version_page_delete'
)
link_document_version_page_list = Link(
    args='resolved_object.pk', icon=icon_document_version_page_list,
    permissions=(permission_document_version_view,), text=_('Pages'),
    view='documents:document_version_page_list'
)
link_document_version_page_list_remap = Link(
    args='resolved_object.pk', icon=icon_document_version_page_list_remap,
    permissions=(permission_document_version_edit,), text=_('Remap pages'),
    view='documents:document_version_page_list_remap'
)
link_document_version_page_navigation_first = Link(
    args='resolved_object.pk', conditional_disable=is_first_page,
    icon=icon_document_version_page_navigation_first,
    keep_query=True, permissions=(permission_document_version_view,),
    text=_('First page'),
    view='documents:document_version_page_navigation_first'
)
link_document_version_page_navigation_last = Link(
    args='resolved_object.pk', conditional_disable=is_last_page,
    icon=icon_document_version_page_navigation_last,
    keep_query=True, text=_('Last page'),
    permissions=(permission_document_version_view,),
    view='documents:document_version_page_navigation_last'
)
link_document_version_page_navigation_previous = Link(
    args='resolved_object.pk', conditional_disable=is_first_page,
    icon=icon_document_version_page_navigation_previous,
    keep_query=True, permissions=(permission_document_version_view,),
    text=_('Previous page'),
    view='documents:document_version_page_navigation_previous'
)
link_document_version_page_navigation_next = Link(
    args='resolved_object.pk', conditional_disable=is_last_page,
    icon=icon_document_version_page_navigation_next,
    keep_query=True, text=_('Next page'),
    permissions=(permission_document_version_view,),
    view='documents:document_version_page_navigation_next'
)
link_document_version_page_return_to_document = Link(
    args='resolved_object.document_version.document.pk',
    icon=icon_document_version_page_return_to_document,
    permissions=(permission_document_view,),
    text=_('Document'), view='documents:document_preview'
)
link_document_version_page_return_to_document_version = Link(
    args='resolved_object.document_version.pk',
    icon=icon_document_version_page_return_to_document_version,
    permissions=(permission_document_version_view,),
    text=_('Document version'), view='documents:document_version_preview'
)
link_document_version_page_return_to_document_version_page_list = Link(
    args='resolved_object.document_version.pk',
    icon=icon_document_version_page_return_to_document_version_page_list,
    permissions=(permission_document_version_view,),
    text=_('Document version pages'),
    view='documents:document_version_page_list'
)
link_document_version_page_rotate_left = Link(
    args='resolved_object.pk', icon=icon_document_version_page_rotate_left,
    keep_query=True, permissions=(permission_document_version_view,),
    text=_('Rotate left'), view='documents:document_version_page_rotate_left'
)
link_document_version_page_rotate_right = Link(
    args='resolved_object.pk', icon=icon_document_version_page_rotate_right,
    keep_query=True, permissions=(permission_document_version_view,),
    text=_('Rotate right'),
    view='documents:document_version_page_rotate_right'
)
link_document_version_page_view = Link(
    args='resolved_object.pk', icon=icon_document_version_page_detail,
    permissions=(permission_document_version_view,), text=_('Page image'),
    view='documents:document_version_page_view'
)
link_document_version_page_view_reset = Link(
    args='resolved_object.pk', icon=icon_document_version_page_detail_reset,
    permissions=(permission_document_version_view,), text=_('Reset view'),
    view='documents:document_version_page_view_reset'
)
link_document_version_page_zoom_in = Link(
    args='resolved_object.pk', conditional_disable=is_max_zoom,
    icon=icon_document_version_page_zoom_in, keep_query=True,
    permissions=(permission_document_version_view,), text=_('Zoom in'),
    view='documents:document_version_page_zoom_in'
)
link_document_version_page_zoom_out = Link(
    args='resolved_object.pk', conditional_disable=is_min_zoom,
    icon=icon_document_version_page_zoom_out, keep_query=True,
    permissions=(permission_document_version_view,), text=_('Zoom out'),
    view='documents:document_version_page_zoom_out'
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0021_auto_20190629_0648')
    ]

    operations = [
        migrations.AddField(
            model_name='imapemail',
            name='execute_expunge',
            field=models.BooleanField(
                default=True, help_text='Execute the IMAP expunge command '
                'after processing each email message.',
                verbose_name='Execute expunge'
            )
        ),
        migrations.AddField(
            model_name='imapemail',
            name='mailbox_destination',
            field=models.CharField(
                blank=True, help_text='IMAP Mailbox to which processed '
                'messages will be copied.', max_length=96, null=True,
                verbose_name='Destination mailbox'
            )
        ),
        migrations.AddField(
            model_name='imapemail',
            name='store_commands',
            field=models.TextField(
                blank=True, default='+FLAGS (\\Deleted)',
                help_text='IMAP STORE command to execute on messages '
                'after they are processed. One command per line. Use '
                'the commands specified in '
                'https://tools.ietf.org/html/rfc2060.html#section-6.4.6 or '
                'the custom commands for your IMAP server.', null=True,
                verbose_name='Store commands'
            )
        ),
        migrations.AddField(
            model_name='imapemail',
            name='search_criteria',
            field=models.TextField(
                blank=True, default='NOT DELETED', help_text='Criteria to '
                'use when searching for messages to process. Use the '
                'format specified in '
                'https://tools.ietf.org/html/rfc2060.html#section-6.4.4',
                null=True, verbose_name='Search criteria'
            ),
        )
    ]



import json

from django.db import migrations


def get_backend_data(self):
    return json.loads(s=self.backend_data or '{}')


def set_backend_data(self, obj):
    self.backend_data = json.dumps(obj=obj)


def code_update_to_credentials(apps, schema_editor):
    UserMailer = apps.get_model(app_label='mailer', model_name='UserMailer')
    StoredCredential = apps.get_model(
        app_label='credentials', model_name='StoredCredential'
    )

    UserMailer.get_backend_data = get_backend_data
    UserMailer.set_backend_data = set_backend_data

    StoredCredential.get_backend_data = get_backend_data
    StoredCredential.set_backend_data = set_backend_data

    queryset = UserMailer.objects.using(
        alias=schema_editor.connection.alias
    ).filter(
        backend_path='mayan.apps.mailer.mailers.DjangoSMTP'
    )

    for obj in queryset:
        obj_backend_data = obj.get_backend_data()

        stored_credential = StoredCredential(
            label='SMTP mailer {} credential'.format(obj.pk),
            internal_name='mailer_smtp_{}_credential'.format(obj.pk),
            backend_path='mayan.apps.credentials.credential_backends.CredentialBackendUsernamePassword'
        )

        stored_credential.set_backend_data(
            obj={
                'password': obj_backend_data['password'],
                'username': obj_backend_data['username']
            }
        )
        stored_credential.save()

        obj_backend_data.update(
            {'stored_credential_id': stored_credential.pk}
        )
        obj_backend_data.pop('password')
        obj_backend_data.pop('username')

        obj.set_backend_data(obj=obj_backend_data)
        obj.save()


def reverse_code_update_to_credentials(apps, schema_editor):
    UserMailer = apps.get_model(app_label='mailer', model_name='UserMailer')
    StoredCredential = apps.get_model(
        app_label='credentials', model_name='StoredCredential'
    )

    UserMailer.get_backend_data = get_backend_data
    UserMailer.set_backend_data = set_backend_data

    StoredCredential.get_backend_data = get_backend_data
    StoredCredential.set_backend_data = set_backend_data

    queryset = UserMailer.objects.using(
        alias=schema_editor.connection.alias
    ).filter(
        backend_path='mayan.apps.mailer.mailers.DjangoSMTP'
    )

    for obj in queryset:
        obj_backend_data = obj.get_backend_data()

        stored_credential = StoredCredential.objects.get(
            pk=obj_backend_data['stored_credential_id']
        )
        credential_backend_data = stored_credential.get_backend_data()

        obj_backend_data.update(
            {
                'password': credential_backend_data['password'],
                'username': credential_backend_data['username']
            }
        )
        obj_backend_data.pop('stored_credential_id')

        obj.set_backend_data(obj=obj_backend_data)
        obj.save()


class Migration(migrations.Migration):
    dependencies = [
        ('credentials', '0001_initial'),
        ('mailer', '0007_auto_20200616_0722'),
    ]

    operations = [
        migrations.RunPython(
            code=code_update_to_credentials,
            reverse_code=reverse_code_update_to_credentials
        )
    ]



from mayan.apps.converter.api_view_mixins import APIImageViewMixin
from mayan.apps.documents.models.document_models import Document
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from .models import SignatureCapture
from .permissions import (
    permission_signature_capture_create, permission_signature_capture_delete,
    permission_signature_capture_edit, permission_signature_capture_view
)
from .serializers import SignatureCaptureSerializer


class APISignatureCaptureDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected signature_capture.
    get: Return the details of the selected signature_capture.
    patch: Edit the selected signature_capture.
    put: Edit the selected signature_capture.
    """
    lookup_url_kwarg = 'signature_capture_id'
    mayan_object_permissions = {
        'DELETE': (permission_signature_capture_delete,),
        'GET': (permission_signature_capture_view,),
        'PATCH': (permission_signature_capture_edit,),
        'PUT': (permission_signature_capture_edit,)
    }
    serializer_class = SignatureCaptureSerializer
    source_queryset = SignatureCapture.valid.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APISignatureCapturesImageView(
    APIImageViewMixin, generics.RetrieveAPIView
):
    """
    get: Returns an image representation of the selected signature capture.
    """
    lookup_url_kwarg = 'signature_capture_id'
    mayan_object_permissions = {
        'GET': (permission_signature_capture_view,)
    }
    source_queryset = SignatureCapture.valid.all()


class APISignatureCaptureListView(
    ExternalObjectAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the signature_captures.
    post: Create a new signature_capture.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    mayan_external_object_permissions = {
        'POST': (permission_signature_capture_create,)
    }
    mayan_object_permissions = {
        'GET': (permission_signature_capture_view,)
    }
    serializer_class = SignatureCaptureSerializer

    def get_instance_extra_data(self):
        return {
            'document': self.get_external_object(),
            'user': self.request.user
        }

    def get_source_queryset(self):
        return self.get_external_object().signature_captures.all()



from django.core import checks
import django.test.runner
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.apps import MayanAppConfig

from .checks import check_app_tests
from .runner import filter_tests_by_tags


class TestingApp(MayanAppConfig):
    name = 'mayan.apps.testing'
    verbose_name = _('Testing')

    def ready(self, *args, **kwargs):
        super().ready(*args, **kwargs)

        checks.register(check=check_app_tests)

        # Replace Django's implementation with the enhanced version.
        setattr(
            django.test.runner, 'filter_tests_by_tags', filter_tests_by_tags
        )



from django.db import migrations, models
import mayan.apps.databases.model_mixins


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('motd', '0005_auto_20160510_0025'),
    ]

    operations = [
        migrations.CreateModel(
            name='Announcement',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True,
                        serialize=False, verbose_name='ID'
                    )
                ),
                (
                    'label', models.CharField(
                        help_text='Short description of this announcement.',
                        max_length=32, verbose_name='Label'
                    )
                ),
                (
                    'text', models.TextField(
                        help_text='The actual test to be displayed.',
                        verbose_name='Text'
                    )
                ),
                (
                    'enabled', models.BooleanField(
                        default=True, verbose_name='Enabled'
                    )
                ),
                (
                    'start_datetime', models.DateTimeField(
                        blank=True, help_text='Date and time after which '
                        'this announcement will be displayed.', null=True,
                        verbose_name='Start date time'
                    )
                ),
                (
                    'end_datetime', models.DateTimeField(
                        blank=True, help_text='Date and time until when '
                        'this announcement is to be displayed.', null=True,
                        verbose_name='End date time'
                    )
                ),
            ],
            options={
                'verbose_name': 'Announcement',
                'verbose_name_plural': 'Announcements',
            },
            bases=(
                mayan.apps.databases.model_mixins.ExtraDataModelMixin,
                models.Model
            ),
        ),
    ]



import yaml

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.serialization import yaml_dump, yaml_load
from mayan.apps.views.forms import DetailForm, ModelForm

from .fields import ImageField
from .models import Asset, LayerTransformation
from .transformations import BaseTransformation


class AssetDetailForm(DetailForm):
    preview = ImageField(
        image_alt_text=_('Asset preview image'), label=_('Preview')
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['preview'].initial = kwargs['instance']

    class Meta:
        fields = ('label', 'internal_name', 'preview')
        model = Asset


class LayerTransformationSelectForm(forms.Form):
    def __init__(self, *args, **kwargs):
        layer = kwargs.pop('layer')
        super().__init__(*args, **kwargs)
        self.fields[
            'transformation'
        ].choices = BaseTransformation.get_transformation_choices(
            layer=layer
        )

    transformation = forms.ChoiceField(
        choices=(), help_text=_('Available transformations for this layer.'),
        label=_('Transformation')
    )


class LayerTransformationForm(ModelForm):
    class Meta:
        fields = ('arguments', 'order')
        model = LayerTransformation

    def __init__(self, *args, **kwargs):
        self._transformation_name = kwargs.pop('transformation_name', None)

        super().__init__(*args, **kwargs)
        transformation_class = self.get_transformation_class()
        if self.instance:
            for key, value in yaml_load(stream=self.instance.arguments or '{}').items():
                self.initial[key] = value

        self.template_name = getattr(
            transformation_class, 'template_name', None
        )

        if self.template_name:
            self.fields['arguments'].widget.attrs['class'] = 'hidden'
            self.fields['order'].widget.attrs['class'] = 'hidden'
        else:
            self.fields['arguments'].widget = forms.widgets.HiddenInput()

    def get_transformation_class(self):
        if not self._transformation_name:
            return self.instance.get_transformation_class()
        else:
            return BaseTransformation.get(name=self._transformation_name)

    def clean(self):
        if self.template_name:
            # If the transformation specifies a template, take the values
            # provided and just check for valid format.
            # Allows compatibility with the redaction template and
            # JavaScript.
            try:
                yaml_load(
                    stream=self.cleaned_data['arguments']
                )
            except yaml.YAMLError:
                raise ValidationError(
                    message=_(
                        '"%s" not a valid entry.'
                    ) % self.cleaned_data['arguments']
                )
        else:
            # Otherwise grab the values from the dynamic form and create
            # the argument JSON object.
            arguments = {}

            for argument in self.get_transformation_class().get_arguments():
                if self.cleaned_data[argument] is not None:
                    arguments[argument] = self.cleaned_data[argument]

            self.cleaned_data['arguments'] = yaml_dump(data=arguments)



from mayan.apps.documents.tests.base import GenericDocumentTestCase

from .mixins import SignatureCaptureTestMixin


class SignatureCaptureModelTestCase(
    SignatureCaptureTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_method_get_absolute_url(self):
        self._create_test_signature_capture()

        self.assertTrue(self._test_signature_capture.get_absolute_url())



from django.utils.translation import ugettext_lazy as _

from .models import StoredCredential
from .permissions import permission_credential_use


class BackendMixinCredentials:
    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'stored_credential_id': {
                    'class': 'mayan.apps.views.fields.FormFieldFilteredModelChoice',
                    'help_text': _(
                        'The credential entry to use for authentication.'
                    ),
                    'kwargs': {
                        'source_model': StoredCredential,
                        'permission': permission_credential_use
                    },
                    'label': _('Credential'),
                    'required': True
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _('Authentication'), {
                    'fields': (
                        'stored_credential_id',
                    )
                },
            ),
        )

        return fieldsets

    def get_credential(self):
        stored_credential_id = self.kwargs.get('stored_credential_id')

        if stored_credential_id:
            return StoredCredential.objects.get(
                pk=stored_credential_id
            ).get_backend_data()


class BackendMixinCredentialsOptional(BackendMixinCredentials):
    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields['stored_credential_id']['required'] = False
        fields['stored_credential_id']['help_text'] = _(
            'Optional credential entry to use for authentication.'
        )

        return fields



from django.template import RequestContext
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.views.generics import (
    SingleObjectCreateView, SingleObjectDeleteView, SingleObjectEditView,
    SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..forms import WorkflowStateEscalationForm
from ..icons import (
    icon_workflow_template_state_escalation_create,
    icon_workflow_template_state_escalation_delete,
    icon_workflow_template_state_escalation_edit,
    icon_workflow_template_state_escalation_list,
    icon_workflow_template_state_escalation
)
from ..links import link_workflow_template_state_escalation_create
from ..models import WorkflowState, WorkflowStateEscalation
from ..permissions import (
    permission_workflow_template_edit, permission_workflow_template_view
)


class WorkflowTemplateStateEscalationCreateView(
    ExternalObjectViewMixin, SingleObjectCreateView
):
    external_object_class = WorkflowState
    external_object_permission = permission_workflow_template_edit
    external_object_pk_url_kwarg = 'workflow_template_state_id'
    form_class = WorkflowStateEscalationForm
    view_icon = icon_workflow_template_state_escalation_create

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'workflow_template_state', 'workflow_template',
            ),
            'title': _(
                'Create escalation for workflow state: %s'
            ) % self.external_object,
            'workflow_template': self.external_object.workflow,
            'workflow_template_state': self.external_object
        }

    def get_form_extra_kwargs(self):
        return {
            'workflow_template_state': self.external_object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'state': self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.escalations.all()

    def get_success_url(self):
        return reverse(
            viewname='document_states:workflow_template_state_escalation_list',
            kwargs={
                'workflow_template_state_id': self.kwargs[
                    'workflow_template_state_id'
                ]
            }
        )


class WorkflowTemplateStateEscalationDeleteView(SingleObjectDeleteView):
    model = WorkflowStateEscalation
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_state_escalation_id'
    view_icon = icon_workflow_template_state_escalation_delete

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow_template_state', 'workflow_template'
            ),
            'object': self.object,
            'title': _('Delete workflow state escalation: %s') % self.object,
            'workflow_template': self.object.state.workflow,
            'workflow_template_state': self.object.state
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_post_escalation_redirect(self):
        return reverse(
            viewname='document_states:workflow_template_state_escalation_list',
            kwargs={
                'workflow_template_state_id': self.object.state.pk
            }
        )


class WorkflowTemplateStateEscalationEditView(SingleObjectEditView):
    form_class = WorkflowStateEscalationForm
    model = WorkflowStateEscalation
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_state_escalation_id'
    view_icon = icon_workflow_template_state_escalation_edit

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow_template_state', 'workflow_template',
            ),
            'object': self.object,
            'title': _('Edit workflow state escalation: %s') % self.object,
            'workflow_template': self.object.state.workflow,
            'workflow_template_state': self.object.state
        }

    def get_form_extra_kwargs(self):
        return {
            'workflow_template_state': self.object.state
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class WorkflowTemplateStateEscalationListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_class = WorkflowState
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_template_state_id'
    view_icon = icon_workflow_template_state_escalation_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'navigation_object_list': (
                'workflow_template_state', 'workflow_template',
            ),
            'no_results_icon': icon_workflow_template_state_escalation,
            'no_results_main_link': link_workflow_template_state_escalation_create.resolve(
                context=RequestContext(
                    dict_={
                        'object': self.external_object
                    }, request=self.request
                )
            ),
            'no_results_text': _(
                'Workflow state escalation allow workflows to execute a '
                'a transition automatically after a specific amount of time.'
            ),
            'no_results_title': _(
                'There are no escalations for this workflow state'
            ),
            'title': _(
                'Escalations for workflow template state: %s'
            ) % self.external_object,
            'workflow_template': self.external_object.workflow,
            'workflow_template_state': self.external_object
        }

    def get_source_queryset(self, queryset=None):
        return self.external_object.escalations.all()



from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import event_workflow_template_edited
from ..models.workflow_state_models import WorkflowState
from ..permissions import (
    permission_workflow_template_edit, permission_workflow_template_view
)

from .literals import (
    TEST_WORKFLOW_TEMPLATE_STATE_LABEL,
    TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION
)
from .mixins.workflow_template_state_mixins import WorkflowTemplateStateViewTestMixin


class WorkflowTemplateStateViewTestCase(
    WorkflowTemplateStateViewTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_workflow_template()

    def test_workflow_state_create_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_template_state_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(WorkflowState.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_create_view_with_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(WorkflowState.objects.count(), 1)
        self.assertEqual(
            WorkflowState.objects.all()[0].label,
            TEST_WORKFLOW_TEMPLATE_STATE_LABEL
        )
        self.assertEqual(
            WorkflowState.objects.all()[0].completion,
            TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)
        self.assertEqual(
            events[0].action_object, self._test_workflow_template_state
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)

    def test_workflow_state_create_invalid_completion_view_with_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_create_view(
            extra_data={'completion': ''}
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(WorkflowState.objects.count(), 1)
        self.assertEqual(
            WorkflowState.objects.all()[0].label,
            TEST_WORKFLOW_TEMPLATE_STATE_LABEL
        )
        self.assertEqual(
            WorkflowState.objects.all()[0].completion, 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)
        self.assertEqual(
            events[0].action_object, self._test_workflow_template_state
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)

    def test_workflow_state_delete_view_no_permission(self):
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()

        self._clear_events()

        response = self._request_test_workflow_template_state_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(WorkflowState.objects.count(), 2)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_delete_view_with_access(self):
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()

        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(WorkflowState.objects.count(), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)
        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)

    def test_workflow_state_edit_view_no_permission(self):
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()

        workflow_state_label = self._test_workflow_template_states[0].label

        self._clear_events()

        response = self._request_test_workflow_template_state_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_workflow_template_states[0].refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_states[0].label, workflow_state_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_edit_view_with_access(self):
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()

        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        workflow_state_label = self._test_workflow_template_states[0].label

        self._clear_events()

        response = self._request_test_workflow_template_state_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_workflow_template_states[0].refresh_from_db()
        self.assertNotEquals(
            self._test_workflow_template_states[0].label, workflow_state_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)
        self.assertEqual(
            events[0].action_object, self._test_workflow_template_states[0]
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)

    def test_workflow_state_list_view_no_permission(self):
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()

        self._clear_events()

        response = self._request_test_workflow_template_state_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_list_view_with_access(self):
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()

        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_list_view()
        self.assertContains(
            response=response, text=self._test_workflow_template_states[0].label,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.conf.urls import url

from .api_views import APICheckedoutDocumentListView, APICheckedoutDocumentView, APIDocumentCheckoutView
from .views import (
    DocumentCheckInView, DocumentCheckOutDetailView, DocumentCheckOutListView,
    DocumentCheckOutView
)

urlpatterns = [
    url(
        regex=r'^documents/$', name='check_out_list',
        view=DocumentCheckOutListView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>\d+)/check/in/$',
        name='check_in_document', view=DocumentCheckInView.as_view()
    ),
    url(
        regex=r'^documents/multiple/check/in/$',
        name='check_in_document_multiple', view=DocumentCheckInView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>\d+)/check/out/$',
        name='check_out_document', view=DocumentCheckOutView.as_view()
    ),
    url(
        regex=r'^documents/multiple/check/out/$',
        name='check_out_document_multiple',
        view=DocumentCheckOutView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>\d+)/checkout/info/$',
        name='check_out_info', view=DocumentCheckOutDetailView.as_view()
    )
]

api_urls = [
    url(
        regex=r'^checkouts/$', name='checkout-document-list',
        view=APICheckedoutDocumentListView.as_view()
    ),
    url(
        regex=r'^checkouts/(?P<checkout_id>[0-9]+)/checkout_info/$',
        name='checkedout-document-view',
        view=APICheckedoutDocumentView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/checkout/$',
        name='document-checkout-view',
        view=APIDocumentCheckoutView.as_view()
    )
]



BINARY_DOCKER_COMPOSE_VERSION = '1.29.2'
DEFAULT_DIRECTORY_BASE = '/opt/'
DEFAULT_DIRECTORY_INSTALLATION = '/opt/mayan-edms/'
DEFAULT_DIRECTORY_MEDIA_ROOT = '/opt/mayan-edms/media/'
DEFAULT_DATABASE_NAME = 'mayan'
DEFAULT_DATABASE_PASSWORD = 'mayanuserpass'
DEFAULT_DATABASE_USER = 'mayan'
DEFAULT_OS_DISTRIBUTION = '"Debian 12"'
DEFAULT_OS_GROUP = 'mayan'
DEFAULT_OS_USERNAME = 'mayan'
DEFAULT_RABBITMQ_PASSWORD = 'mayanrabbitmqpassword'
DEFAULT_RABBITMQ_USER = 'mayan'
DEFAULT_RABBITMQ_VHOST = 'mayan'
DEFAULT_REDIS_PASSWORD = 'mayanredispassword'
DEFAULT_SECRET_KEY = 'secret-key-missing!'
DEFAULT_USER_SETTINGS_FOLDER = 'user_settings/'
DEFAULT_USER_SETTINGS_MODULE = 'user_settings'
DOCKER_HOST_REGISTRY_NAME = 'docker-registry.local'
DOCKER_HOST_REGISTRY_PORT = 5000
DOCKER_IMAGE_MAYAN_NAME = 'mayanedms/mayanedms'
DOCKER_DIND_IMAGE_VERSION = 'docker:23.0.6-dind'
DOCKER_ELASTIC_IMAGE_VERSION = 'elasticsearch:7.17.9'
DOCKER_KEYCLOAK_IMAGE_VERSION = 'keycloak/keycloak:20.0.1'
DOCKER_LINUX_IMAGE_VERSION = 'debian:12.1-slim'
DOCKER_MYSQL_IMAGE_VERSION = 'mysql:8.0.34'
DOCKER_ORACLE_IMAGE_VERSION = 'wnameless/oracle-xe-11g-r2'
DOCKER_POSTGRES_IMAGE_VERSION = 'postgres:13.11-alpine'
DOCKER_PYTHON_IMAGE_VERSION = 'python:3.11.4-slim'
DOCKER_RABBITMQ_IMAGE_VERSION = 'rabbitmq:3.12.2-alpine'
DOCKER_REDIS_IMAGE_VERSION = 'redis:7.0.12-alpine'
DOCKER_TRAEFIK_IMAGE_VERSION = 'traefik:v2.5'
GIT_REMOTE_NAME = 'origin'
GITLAB_CI_BRANCH_BUILDS_DOCKER = 'builds/docker'
GITLAB_CI_BRANCH_BUILDS_DOCUMENTATION = 'builds/documentation'
GITLAB_CI_BRANCH_BUILDS_PYTHON = 'builds/python'
GITLAB_CI_BRANCH_DEPLOYMENTS_DEMO = 'deployments/demo'
GITLAB_CI_BRANCH_DEPLOYMENTS_STAGING = 'deployments/staging'
GITLAB_CI_BRANCH_RELEASES_ALL_MAJOR = 'releases/all/major'
GITLAB_CI_BRANCH_RELEASES_ALL_MINOR = 'releases/all/minor'
GITLAB_CI_BRANCH_RELEASES_DOCKER_MAJOR = 'releases/docker/major'
GITLAB_CI_BRANCH_RELEASES_DOCKER_MINOR = 'releases/docker/minor'
GITLAB_CI_BRANCH_RELEASES_DOCUMENTATION = 'releases/documentation'
GITLAB_CI_BRANCH_RELEASES_NIGHTLY = 'releases/nightly'
GITLAB_CI_BRANCH_RELEASES_PYTHON_MAJOR = 'releases/python/major'
GITLAB_CI_BRANCH_RELEASES_PYTHON_MINOR = 'releases/python/minor'
GITLAB_CI_BRANCH_RELEASES_STAGING = 'releases/staging'
GITLAB_CI_BRANCH_RELEASES_TESTING = 'releases/testing'
GITLAB_CI_BRANCH_TESTS_ALL = 'tests/all'
GITLAB_CI_BRANCH_TESTS_DOCKER = 'tests/docker'
GITLAB_CI_BRANCH_TESTS_PYTHON_ALL = 'tests/python/all'
GITLAB_CI_BRANCH_TESTS_PYTHON_BASE = 'tests/python/base'
GITLAB_CI_BRANCH_TESTS_PYTHON_UPGRADE = 'tests/python/upgrade'
GUNICORN_LIMIT_REQUEST_LINE = 4094
GUNICORN_MAX_REQUESTS = 500
GUNICORN_REQUESTS_JITTER = 50
GUNICORN_TIMEOUT = 120
GUNICORN_WORKER_CLASS = 'sync'
GUNICORN_WORKERS = 3
LOCAL_VERSION = ''
PYTHON_AMQP_VERSION = '5.1.0'
PYTHON_MYSQL_VERSION = '2.0.3'
PYTHON_ORACLE_VERSION = '8.1.0'
PYTHON_PIP_VERSION = '23.2.1'
PYTHON_PSYCOPG2_VERSION = '2.9.6'
PYTHON_PSYCOPG2_VERSION_PREVIOUS = '2.9.3'
PYTHON_PSUTIL_VERSION = '5.8.0'
PYTHON_REDIS_VERSION = '4.6.0'
PYTHON_WHEEL_VERSION = '0.41.0'
SOURCE_CODE_REPOSITORY = 'https://gitlab.com/mayan-edms/mayan-edms/'
SOURCE_CODE_GIT = 'https://gitlab.com/mayan-edms/mayan-edms.git'
SOURCE_CODE_ISSUES = 'https://gitlab.com/mayan-edms/mayan-edms/issues/'
SECRET_KEY_FILENAME = 'SECRET_KEY'
SYSTEM_DIR = 'system'
SUPERVISOR_CONFIGURATION_DIRECTORY = '/etc/supervisor/conf.d/'
SUPERVISOR_CONFIGURATION_FILENAME = 'mayan-edms.conf'
MAYAN_WORKER_A_CONCURRENCY = 0
MAYAN_WORKER_A_MAX_MEMORY_PER_CHILD = 400000
MAYAN_WORKER_A_MAX_TASKS_PER_CHILD = 1000
MAYAN_WORKER_B_CONCURRENCY = 0
MAYAN_WORKER_B_MAX_MEMORY_PER_CHILD = 400000
MAYAN_WORKER_B_MAX_TASKS_PER_CHILD = 1000
MAYAN_WORKER_C_CONCURRENCY = 0
MAYAN_WORKER_C_MAX_MEMORY_PER_CHILD = 400000
MAYAN_WORKER_C_MAX_TASKS_PER_CHILD = 1000
MAYAN_WORKER_D_CONCURRENCY = 1
MAYAN_WORKER_D_MAX_MEMORY_PER_CHILD = 400000
MAYAN_WORKER_D_MAX_TASKS_PER_CHILD = 100
MAYAN_WORKER_E_CONCURRENCY = 0
MAYAN_WORKER_E_MAX_MEMORY_PER_CHILD = 400000
MAYAN_WORKER_E_MAX_TASKS_PER_CHILD = 1000



TEST_SOURCE_BACKEND_PATH_STORAGE = 'mayan.apps.source_stored_files.tests.source_backends.SourceBackendTestStorage'

TEST_STAGING_PREVIEW_MAX_SIZE = 100000000
TEST_STAGING_PREVIEW_HEIGHT = ''
TEST_STAGING_PREVIEW_WIDTH = 640



import pytz

from django.conf import settings
from django.db import migrations, reset_queries


def code_document_version_page_create(apps, schema_editor):
    ContentType = apps.get_model(
        app_label='contenttypes', model_name='ContentType'
    )
    DocumentFilePage = apps.get_model(
        app_label='documents', model_name='DocumentFilePage'
    )
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )

    content_type = ContentType.objects.get_for_model(model=DocumentFilePage)
    content_type_id = content_type.id
    cursor_main = schema_editor.connection.create_cursor(name='document_file_to_version')
    cursor_document_version_page = schema_editor.connection.cursor()

    query = '''
    SELECT
        {documents_documentfile}.{document_id},
        {documents_documentfile}.{timestamp},
        {documents_documentfilepage}.{document_file_id},
        {documents_documentfilepage}.{id}
    FROM {documents_documentfilepage}
    INNER JOIN {documents_documentfile} ON (
        {documents_documentfilepage}.{document_file_id} = {documents_documentfile}.{id}
    ) ORDER BY {documents_documentfilepage}.{id} ASC
    '''.format(
        documents_documentfile=schema_editor.connection.ops.quote_name(
            name='documents_documentfile'
        ),
        document_id=schema_editor.connection.ops.quote_name(
            name='document_id'
        ),
        timestamp=schema_editor.connection.ops.quote_name(name='timestamp'),
        documents_documentfilepage=schema_editor.connection.ops.quote_name(
            name='documents_documentfilepage'
        ),
        document_file_id=schema_editor.connection.ops.quote_name(
            name='document_file_id'
        ),
        id=schema_editor.connection.ops.quote_name(name='id')
    )

    cursor_main.execute(query)

    class DummyDocumentVersion:
        def save(self):
            """Does not do anything."""

    document_file_id_last = None
    document_version = DummyDocumentVersion()

    document_version_page_insert_query = '''
        INSERT INTO {documents_documentversionpage} (
            {document_version_id},{content_type_id},{object_id},{page_number}
        ) VALUES {{}};
    '''.format(
        content_type_id=schema_editor.connection.ops.quote_name(
            name='content_type_id'
        ),
        document_version_id=schema_editor.connection.ops.quote_name(
            name='document_version_id'
        ),
        documents_documentversionpage=schema_editor.connection.ops.quote_name(
            name='documents_documentversionpage'
        ),
        object_id=schema_editor.connection.ops.quote_name(
            name='object_id'
        ),
        page_number=schema_editor.connection.ops.quote_name(
            name='page_number'
        )
    )

    document_version_page_values = []
    page_number = 1

    FETCH_SIZE = 100000

    time_zone = pytz.timezone(zone=settings.TIME_ZONE)

    while True:
        rows = cursor_main.fetchmany(FETCH_SIZE)

        if not rows:
            break

        for row in rows:
            document_id, timestamp, document_file_id, document_file_page_id = row
            timestamp = timestamp.astimezone(tz=time_zone)

            if document_file_id_last != document_file_id:
                document_version = DocumentVersion.objects.create(
                    document_id=document_id
                )
                document_version.timestamp = timestamp
                document_version.save()
                document_version_id = document_version.pk
                document_file_id_last = document_file_id
                if document_version_page_values:
                    final_query = document_version_page_insert_query.format(
                        ('(%s,%s,%s,%s),' * (page_number - 1))[:-1]
                    )

                    page_number = 1

                    cursor_document_version_page.execute(
                        final_query, document_version_page_values
                    )
                    reset_queries()

                document_version_page_values = []

            document_version_page_values += (
                document_version_id, content_type_id, document_file_page_id,
                page_number
            )

            page_number += 1

    if page_number > 1:
        final_query = document_version_page_insert_query.format(
            ('(%s,%s,%s,%s),' * (page_number - 1))[:-1]
        )
        cursor_document_version_page.execute(
            final_query, document_version_page_values
        )


def code_document_version_page_create_reverse(apps, schema_editor):
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )

    DocumentVersion.objects.using(
        alias=schema_editor.connection.alias
    ).all().delete()


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('documents', '0062_auto_20200920_0614')
    ]

    operations = [
        migrations.RunPython(
            code=code_document_version_page_create,
            reverse_code=code_document_version_page_create_reverse
        )
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        (
            'document_states',
            '0025_alter_workflowtransitiontriggerevent_unique_together'
        )
    ]

    operations = [
        migrations.AlterModelOptions(
            name='workflowtransitiontriggerevent',
            options={
                'ordering': ('event_type__name',),
                'verbose_name': 'Workflow transition trigger event',
                'verbose_name_plural': 'Workflow transitions trigger events'
            }
        )
    ]



from django.db.models import Q

from ...models.workflow_transition_models import WorkflowTransition

from ..literals import (
    TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED
)

from .workflow_template_state_mixins import WorkflowTemplateStateTestMixin


class WorkflowTemplateTransitionTestMixin(WorkflowTemplateStateTestMixin):
    def setUp(self):
        super().setUp()
        self._test_workflow_template_transitions = []

    def _create_test_workflow_template_transition(self, extra_kwargs=None):
        total_test_workflow_template_transitions = len(
            self._test_workflow_template_transitions
        )
        label = '{}_{}'.format(
            TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
            total_test_workflow_template_transitions
        )

        kwargs = {
            'label': label,
            'origin_state': self._test_workflow_template_states[0],
            'destination_state': self._test_workflow_template_states[1]
        }

        if extra_kwargs is not None:
            kwargs.update(extra_kwargs)

        self._test_workflow_template_transition = self._test_workflow_template.transitions.create(
            **kwargs
        )

        self._test_workflow_template_transitions.append(
            self._test_workflow_template_transition
        )


class WorkflowTemplateTransitionAPIViewTestMixin(
    WorkflowTemplateTransitionTestMixin
):
    def _request_test_workflow_template_transition_create_api_view(
        self, extra_data=None
    ):
        data = {
            'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
            'origin_state_id': self._test_workflow_template_states[0].pk,
            'destination_state_id': self._test_workflow_template_states[1].pk
        }

        if extra_data:
            data.update(extra_data)

        pk_list = list(
            WorkflowTransition.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='rest_api:workflow-template-transition-list', kwargs={
                'workflow_template_id': self._test_workflow_template.pk
            }, data=data
        )

        try:
            self._test_workflow_template_transition = WorkflowTransition.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowTransition.DoesNotExist:
            self._test_workflow_template_transition = None

        return response

    def _request_test_workflow_template_transition_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )

    def _request_test_workflow_template_transition_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )

    def _request_test_workflow_template_transition_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_transition_edit_patch_api_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED,
                'origin_state_id': self._test_workflow_template_states[1].pk,
                'destination_state_id': self._test_workflow_template_states[0].pk
            }
        )

    def _request_test_workflow_template_transition_edit_put_api_view_via(self):
        return self.put(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED,
                'origin_state_id': self._test_workflow_template_states[1].pk,
                'destination_state_id': self._test_workflow_template_states[0].pk
            }
        )


class WorkflowTemplateTransitionViewTestMixin(
    WorkflowTemplateTransitionTestMixin
):
    def _request_test_workflow_template_transition_create_view(self):
        pk_list = list(
            WorkflowTransition.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='document_states:workflow_template_transition_create',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
                'origin_state': self._test_workflow_template_states[0].pk,
                'destination_state': self._test_workflow_template_states[1].pk
            }
        )

        try:
            self._test_workflow_template_transition = WorkflowTransition.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowTransition.DoesNotExist:
            self._test_workflow_template_transition = None

        return response

    def _request_test_workflow_template_transition_delete_view(self):
        return self.post(
            viewname='document_states:workflow_template_transition_delete',
            kwargs={
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )

    def _request_test_workflow_template_transition_edit_view(self):
        return self.post(
            viewname='document_states:workflow_template_transition_edit',
            kwargs={
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED,
                'origin_state': self._test_workflow_template_states[0].pk,
                'destination_state': self._test_workflow_template_states[1].pk
            }
        )

    def _request_test_workflow_template_transition_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_transition_list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk
            }
        )



from django.apps import apps
from django.utils.translation import ugettext_lazy as _

from mayan.apps.document_indexing.tasks import task_index_instance_document_add
from mayan.apps.events.classes import EventType
from mayan.apps.events.literals import TEXT_UNKNOWN_EVENT_ID

from .literals import STORAGE_NAME_WORKFLOW_CACHE
from .settings import setting_workflow_image_cache_maximum_size
from .tasks import task_launch_all_workflow_for


def handler_create_workflow_image_cache(sender, **kwargs):
    Cache = apps.get_model(app_label='file_caching', model_name='Cache')
    Cache.objects.update_or_create(
        defaults={
            'maximum_size': setting_workflow_image_cache_maximum_size.value,
        }, defined_storage_name=STORAGE_NAME_WORKFLOW_CACHE,
    )


def handler_launch_workflow_on_create(sender, instance, created, **kwargs):
    if created:
        task_launch_all_workflow_for.apply_async(
            kwargs={'document_id': instance.pk}
        )


def handler_launch_workflow_on_type_change(sender, instance, **kwargs):
    task_launch_all_workflow_for.apply_async(
        kwargs={'document_id': instance.pk}
    )


def handler_trigger_transition(sender, **kwargs):
    action = kwargs['instance']

    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )
    WorkflowInstance = apps.get_model(
        app_label='document_states', model_name='WorkflowInstance'
    )
    WorkflowTransition = apps.get_model(
        app_label='document_states', model_name='WorkflowTransition'
    )

    trigger_transitions = WorkflowTransition.objects.filter(
        trigger_events__event_type__name=kwargs['instance'].verb
    )

    if isinstance(action.target, Document):
        workflow_instances = WorkflowInstance.objects.filter(
            workflow__transitions__in=trigger_transitions,
            document=action.target
        ).distinct()
    elif isinstance(action.action_object, Document):
        workflow_instances = WorkflowInstance.objects.filter(
            workflow__transitions__in=trigger_transitions,
            document=action.action_object
        ).distinct()
    else:
        workflow_instances = WorkflowInstance.objects.none()

    for workflow_instance in workflow_instances:
        # Select the first transition that is valid for this workflow state.
        valid_transitions = list(
            set(
                trigger_transitions
            ) & set(
                workflow_instance.get_transition_choices()
            )
        )
        if valid_transitions:
            try:
                event_type_label = EventType.get(id=action.verb).label
            except KeyError:
                event_type_label = TEXT_UNKNOWN_EVENT_ID % action.verb

            workflow_instance.do_transition(
                comment=_('Event trigger: %s') % event_type_label,
                transition=valid_transitions[0]
            )


# Indexing, workflow template


def handler_workflow_template_post_edit(sender, **kwargs):
    if not kwargs.get('created', False):
        for workflow_instance in kwargs['instance'].instances.all():
            task_index_instance_document_add.apply_async(
                kwargs={'document_id': workflow_instance.document.pk}
            )


# Indexing, workflow state


def handler_workflow_template_state_post_edit(sender, **kwargs):
    if not kwargs.get('created', False):
        for workflow_instance in kwargs['instance'].workflow.instances.all():
            task_index_instance_document_add.apply_async(
                kwargs={'document_id': workflow_instance.document.pk}
            )


def handler_workflow_template_state_pre_delete(sender, **kwargs):
    for workflow_instance in kwargs['instance'].workflow.instances.all():
        # Remove each of the documents.
        # Trigger the remove event for each document so they can be
        # reindexed.
        workflow_instance.delete()
        task_index_instance_document_add.apply_async(
            kwargs={'document_id': workflow_instance.document.pk}
        )


# Indexing, workflow template


def handler_workflow_template_transition_post_edit(sender, **kwargs):
    if not kwargs.get('created', False):
        for workflow_instance in kwargs['instance'].workflow.instances.all():
            task_index_instance_document_add.apply_async(
                kwargs={'document_id': workflow_instance.document.pk}
            )


def handler_workflow_template_transition_pre_delete(sender, **kwargs):
    for workflow_instance in kwargs['instance'].workflow.instances.all():
        # Remove each of the documents.
        # Trigger the remove event for each document so they can be
        # reindexed.
        workflow_instance.delete()
        task_index_instance_document_add.apply_async(
            kwargs={'document_id': workflow_instance.document.pk}
        )



from mayan.apps.django_gpg.permissions import permission_key_sign
from mayan.apps.django_gpg.tests.mixins import KeyTestMixin
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.literals import (
    TEST_FILE_PDF_PATH, TEST_FILE_SMALL_PATH
)

from ..events import (
    event_detached_signature_created, event_detached_signature_deleted,
    event_detached_signature_uploaded
)
from ..models import DetachedSignature
from ..permissions import (
    permission_document_file_sign_detached,
    permission_document_file_signature_delete,
    permission_document_file_signature_download,
    permission_document_file_signature_upload,
    permission_document_file_signature_view
)

from .mixins import (
    DetachedSignatureViewTestMixin, DetachedSignatureTestMixin
)


class DetachedSignaturesViewTestCase(
    KeyTestMixin, DetachedSignatureTestMixin, DetachedSignatureViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_detached_signature_create_view_with_no_permission(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self._test_document.file_latest.signatures.count()

        self._clear_events()

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signature_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_detached_signature_create_view_with_document_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_detached
        )

        self._clear_events()

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signature_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_detached_signature_create_view_with_key_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._clear_events()

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signature_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_detached_signature_create_view_with_full_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_detached
        )
        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._clear_events()

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signature_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object,
            self._test_document_file.signatures.first().detachedsignature
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(events[0].verb, event_detached_signature_created.id)

    def test_trashed_document_detached_signature_create_view_with_full_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_detached
        )
        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signature_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_detached_signature_delete_view_no_permission(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature_count = DetachedSignature.objects.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self._clear_events()

        response = self._request_test_document_file_signature_detached_delete_view()
        self.assertEqual(response.status_code, 404)
        self.assertEqual(DetachedSignature.objects.count(), signature_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_detached_signature_delete_view_with_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature_count = DetachedSignature.objects.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_delete
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self._clear_events()

        response = self._request_test_document_file_signature_detached_delete_view()
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            DetachedSignature.objects.count(), signature_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_document_file)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(events[0].verb, event_detached_signature_deleted.id)

    def test_trashed_document_signature_delete_view_with_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature_count = DetachedSignature.objects.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_delete
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_signature_detached_delete_view()
        self.assertEqual(response.status_code, 404)
        self.assertEqual(
            DetachedSignature.objects.count(), signature_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_download_view_no_permission(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        self._clear_events()

        response = self._request_test_document_file_signature_detached_download_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_download_view_with_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_download
        )

        self.expected_content_types = ('application/octet-stream',)

        self._clear_events()

        response = self._request_test_document_file_signature_detached_download_view()

        with self._test_signature.signature_file as file_object:
            self.assert_download_response(
                response=response, content=file_object.read(),
            )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_download_view_with_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_download
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_signature_detached_download_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_upload_view_no_permission(self):
        self._test_document_path = TEST_FILE_PDF_PATH

        signature_count = DetachedSignature.objects.count()

        self._upload_test_document()

        self._clear_events()

        response = self._request_test_document_file_signature_detached_upload_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(DetachedSignature.objects.count(), signature_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_upload_view_with_access(self):
        self._test_document_path = TEST_FILE_PDF_PATH
        self._upload_test_document()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_upload
        )

        signature_count = DetachedSignature.objects.count()

        self._clear_events()

        response = self._request_test_document_file_signature_detached_upload_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            DetachedSignature.objects.count(), signature_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object,
            self._test_document.file_latest.signatures.first().detachedsignature
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(events[0].verb, event_detached_signature_uploaded.id)

    def test_trashed_document_signature_upload_view_with_access(self):
        self._test_document_path = TEST_FILE_PDF_PATH
        self._upload_test_document()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_upload
        )

        self._test_document.delete()

        signature_count = DetachedSignature.objects.count()

        self._clear_events()

        response = self._request_test_document_file_signature_detached_upload_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(DetachedSignature.objects.count(), signature_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_('Mailing'), name='mailing'
)

event_email_sent = namespace.add_event_type(
    label=_('Email sent'), name='email_send'
)
event_profile_created = namespace.add_event_type(
    label=_('Mailing profile created'), name='profile_created'
)
event_profile_edited = namespace.add_event_type(
    label=_('Mailing profile edited'), name='profile_edited'
)



from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.views.generics import (
    MultipleObjectDeleteView, SingleObjectDownloadView, SingleObjectListView
)

from .icons import (
    icon_download_file_delete, icon_download_file_download,
    icon_download_file_list
)
from .models import DownloadFile
from .permissions import (
    permission_download_file_delete, permission_download_file_download,
    permission_download_file_view
)
from .view_mixins import OwnerPlusFilteredQuerysetViewMixin


class DownloadFileDeleteView(
    OwnerPlusFilteredQuerysetViewMixin, MultipleObjectDeleteView
):
    model = DownloadFile
    optional_object_permission = permission_download_file_delete
    pk_url_kwarg = 'download_file_id'
    post_action_redirect = reverse_lazy(
        viewname='storage:download_file_list'
    )
    view_icon = icon_download_file_delete

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class DownloadFileDownloadViewView(
    OwnerPlusFilteredQuerysetViewMixin, SingleObjectDownloadView
):
    model = DownloadFile
    optional_object_permission = permission_download_file_download
    pk_url_kwarg = 'download_file_id'
    view_icon = icon_download_file_download

    def get_download_file_object(self):
        instance = self.get_object()
        instance._event_actor = self.request.user
        return instance.get_download_file_object()

    def get_download_filename(self):
        return self.object.filename or str(self.object)


class DownloadFileListView(
    OwnerPlusFilteredQuerysetViewMixin, SingleObjectListView
):
    model = DownloadFile
    optional_object_permission = permission_download_file_view
    view_icon = icon_download_file_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_download_file_list,
            'no_results_text': _(
                'Download files are created as a results of a an external '
                'process like an export. Download files are retained over '
                'a span of time and then removed automatically.'
            ),
            'no_results_title': _('There are no files to download.'),
            'title': _('Downloads')
        }



from ..literals import TEST_SOURCE_BACKEND_PATH_DOCUMENT_BASIC

from .base_mixins import SourceTestMixin


class SourceDocumentUploadWizardTestMixin(SourceTestMixin):
    _test_source_backend_path = TEST_SOURCE_BACKEND_PATH_DOCUMENT_BASIC

    def _request_document_upload_wizard_get_view(self):
        return self.get(viewname='sources:document_upload_wizard')

    def _request_document_upload_wizard_post_view(self):
        # Get the step after the document type selection.
        return self.post(
            viewname='sources:document_upload_wizard', data={
                'document_create_wizard-current_step': 0,
                'document_type_selection-document_type': self._test_document_type.pk
            }
        )



from django.urls import reverse

from ..links.document_file_links import link_document_file_delete
from ..links.favorite_links import (
    link_document_favorites_add, link_document_favorites_remove
)
from ..links.trashed_document_links import link_document_restore
from ..models import TrashedDocument
from ..permissions import (
    permission_document_file_delete, permission_document_view,
    permission_trashed_document_restore
)

from .base import GenericDocumentViewTestCase
from .mixins.document_file_mixins import DocumentFileTestMixin
from .mixins.favorite_document_mixins import FavoriteDocumentTestMixin


class FavoriteDocumentLinkTestCase(
    FavoriteDocumentTestMixin, GenericDocumentViewTestCase
):
    auto_create_test_document_stub = True
    auto_upload_test_document = False

    def test_favorite_document_add_link_no_permission(self):
        self.add_test_view(test_object=self._test_document)
        context = self.get_test_view()
        resolved_link = link_document_favorites_add.resolve(context=context)

        self.assertEqual(resolved_link, None)

    def test_favorite_document_add_link_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self.add_test_view(test_object=self._test_document)
        context = self.get_test_view()
        resolved_link = link_document_favorites_add.resolve(context=context)

        self.assertNotEqual(resolved_link, None)

    def test_favorite_document_add_link_external_user_with_access(self):
        self._create_test_user()
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._test_document_favorite_add(user=self._test_user)

        self.add_test_view(test_object=self._test_document)
        context = self.get_test_view()
        resolved_link = link_document_favorites_add.resolve(context=context)

        self.assertNotEqual(resolved_link, None)

    def test_favorite_document_remove_link_no_permission(self):
        self._test_document_favorite_add()

        self.add_test_view(test_object=self._test_document)
        context = self.get_test_view()
        resolved_link = link_document_favorites_remove.resolve(
            context=context
        )

        self.assertEqual(resolved_link, None)

    def test_favorite_document_remove_link_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._test_document_favorite_add()

        self.add_test_view(test_object=self._test_document)
        context = self.get_test_view()
        resolved_link = link_document_favorites_remove.resolve(
            context=context
        )

        self.assertNotEqual(resolved_link, None)

    def test_favorite_document_remove_link_external_user_with_access(self):
        self._create_test_user()
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._test_document_favorite_add(user=self._test_user)

        self.add_test_view(test_object=self._test_document)
        context = self.get_test_view()
        resolved_link = link_document_favorites_remove.resolve(
            context=context
        )

        self.assertEqual(resolved_link, None)


class DocumentsLinksTestCase(
    DocumentFileTestMixin, GenericDocumentViewTestCase
):
    def test_document_file_delete_link_no_permission(self):
        self._upload_test_document_file()

        self.assertEqual(
            self._test_document.files.count(), 2
        )

        self.add_test_view(
            test_object=self._test_document.files.first()
        )
        context = self.get_test_view()
        resolved_link = link_document_file_delete.resolve(context=context)

        self.assertEqual(resolved_link, None)

    def test_document_file_delete_link_with_permission(self):
        self._upload_test_document_file()

        self.assertEqual(
            self._test_document.files.count(), 2
        )

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_delete
        )

        self.add_test_view(
            test_object=self._test_document.files.first()
        )
        context = self.get_test_view()
        resolved_link = link_document_file_delete.resolve(context=context)

        self.assertNotEqual(resolved_link, None)
        self.assertEqual(
            resolved_link.url,
            reverse(
                viewname=link_document_file_delete.view,
                args=(
                    self._test_document.files.first().pk,
                )
            )
        )


class TrashedDocumentsLinksTestCase(GenericDocumentViewTestCase):
    def setUp(self):
        super().setUp()
        self._test_document.delete()
        self.test_trashed_document = TrashedDocument.objects.get(
            pk=self._test_document.pk
        )
        self.add_test_view(test_object=self.test_trashed_document)
        self.context = self.get_test_view()

    def test_trashed_document_restore_link_no_permission(self):
        resolved_link = link_document_restore.resolve(context=self.context)
        self.assertEqual(resolved_link, None)

    def test_trashed_document_restore_link_with_permission(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_trashed_document_restore
        )
        resolved_link = link_document_restore.resolve(context=self.context)
        self.assertNotEqual(resolved_link, None)
        self.assertEqual(
            resolved_link.url,
            reverse(
                viewname=link_document_restore.view,
                args=(self.test_trashed_document.pk,)
            )
        )



from pathlib import Path as OriginalPath


class Path(type(OriginalPath())):
    """
    TODO: Remove this class when the minimum required Python version
    becomes 3.9.
    """

    def is_relative_to(self, *other):
        """
        Backported method for Python < 3.9.
        Return whether or not this path is relative to the other path.
        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.is_relative_to
        """
        try:
            self.relative_to(*other)
        except ValueError:
            return False
        else:
            return True



from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_page_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.literals import TEST_FILE_COMPRESSED_PATH
from mayan.apps.source_compressed.source_backends.literals import (
    SOURCE_UNCOMPRESS_CHOICE_ALWAYS, SOURCE_UNCOMPRESS_CHOICE_NEVER
)
from mayan.apps.sources.events import event_source_created
from mayan.apps.sources.models import Source

from mayan.apps.sources.permissions import (
    permission_sources_create, permission_sources_edit
)
from mayan.apps.sources.tests.mixins.source_view_mixins import SourceViewTestMixin

from .mixins import WatchFolderSourceTestMixin


class WatchFolderSourceBackendViewTestCase(
    WatchFolderSourceTestMixin, SourceViewTestMixin,
    GenericDocumentViewTestCase
):
    _test_source_create_auto = False
    auto_upload_test_document = False

    def test_watch_folder_source_create_get_view_no_permission(self):
        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_get_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_watch_folder_source_create_get_view_with_permission(self):
        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_get_view()

        self.assertEqual(response.status_code, 200)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_watch_folder_source_create_post_view_no_permission(self):
        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_post_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_watch_folder_source_create_post_view_with_permission(self):
        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Source.objects.count(), source_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_source)
        self.assertEqual(events[0].verb, event_source_created.id)

    def test_watch_folder_source_test_view_no_permission(self):
        self._test_source_create()
        self.copy_test_source_file()

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.objects.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_watch_folder_source_test_view_with_access(self):
        self._test_source_create()
        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.objects.count(), document_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )


class WatchFolderSourceTestViewTestCase(
    WatchFolderSourceTestMixin, SourceViewTestMixin,
    GenericDocumentViewTestCase
):
    _test_source_create_auto = False
    auto_upload_test_document = False

    def test_source_test_get_view_no_permission(self):
        self._test_source_create()

        self.copy_test_source_file()

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_test_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.objects.count(), document_count)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_test_get_view_with_access(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_test_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(Document.objects.count(), document_count)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_test_post_view_no_permission(self):
        self._test_source_create()

        self.copy_test_source_file()

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.objects.count(), document_count)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_test_post_view_with_access(self):
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.objects.count(), document_count + 1)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

    def test_source_compressed_always_test_post_view_with_access(self):
        self._test_source_create(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_ALWAYS}
        )

        self.copy_test_source_file(source_path=TEST_FILE_COMPRESSED_PATH)

        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.objects.count(), document_count + 2)

        self.assertTrue(
            'first document.pdf' in Document.objects.values_list(
                'label', flat=True
            )
        )
        self.assertTrue(
            'second document.pdf' in Document.objects.values_list(
                'label', flat=True
            )
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 11)

        test_documents = Document.objects.all()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_documents[0])
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_documents[0])
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_documents[0].file_latest)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_documents[0])
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_documents[0].file_latest)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_documents[0])
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_documents[0].version_active)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, test_documents[0].version_active
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(
            events[4].target, test_documents[0].version_active.pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_document_type)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_documents[1])
        self.assertEqual(events[5].verb, event_document_created.id)

        self.assertEqual(events[6].action_object, test_documents[1])
        self.assertEqual(events[6].actor, self._test_case_user)
        self.assertEqual(events[6].target, test_documents[1].file_latest)
        self.assertEqual(events[6].verb, event_document_file_created.id)

        self.assertEqual(events[7].action_object, test_documents[1])
        self.assertEqual(events[7].actor, self._test_case_user)
        self.assertEqual(events[7].target, test_documents[1].file_latest)
        self.assertEqual(events[7].verb, event_document_file_edited.id)

        self.assertEqual(events[8].action_object, test_documents[1])
        self.assertEqual(events[8].actor, self._test_case_user)
        self.assertEqual(events[8].target, test_documents[1].version_active)
        self.assertEqual(events[8].verb, event_document_version_created.id)

        self.assertEqual(
            events[9].action_object, test_documents[1].version_active
        )
        self.assertEqual(events[9].actor, self._test_case_user)
        self.assertEqual(
            events[9].target, test_documents[1].version_active.pages.first()
        )
        self.assertEqual(
            events[9].verb, event_document_version_page_created.id
        )

        self.assertEqual(
            events[10].action_object, test_documents[1].version_active
        )
        self.assertEqual(events[10].actor, self._test_case_user)
        self.assertEqual(
            events[10].target, test_documents[1].version_active.pages.last()
        )
        self.assertEqual(
            events[10].verb, event_document_version_page_created.id
        )

    def test_source_compressed_never_test_post_view_with_access(self):
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_source_create(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER}
        )

        self.copy_test_source_file(source_path=TEST_FILE_COMPRESSED_PATH)

        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.objects.count(), document_count + 1)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

    def test_source_compressed_always_non_compressed_test_post_view_with_access(self):
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_source_create(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_ALWAYS}
        )

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.objects.count(), document_count + 1)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )



from django.contrib import messages
from django.utils.translation import ugettext_lazy as _

from actstream.models import Action

from mayan.apps.databases.classes import QuerysetParametersSerializer
from mayan.apps.views.generics import ConfirmView
from mayan.apps.views.view_mixins import ExternalContentTypeObjectViewMixin

from ..icons import (
    icon_event_list_export, icon_object_event_list_export,
    icon_verb_event_list_export
)
from ..permissions import permission_events_export
from ..tasks import task_event_queryset_export

from .view_mixins import VerbEventViewMixin


class EventExportBaseView(ConfirmView):
    object_permission = permission_events_export
    view_icon = icon_event_list_export

    def get_extra_context(self):
        return {
            'message': _(
                'The process will be performed in the background. '
                'The exported events will be available in the downloads '
                'area.'
            )
        }

    def view_action(self):
        decomposed_queryset = QuerysetParametersSerializer.decompose(
            _model=Action, **self.get_queryset_parameters()
        )

        task_event_queryset_export.apply_async(
            kwargs={
                'decomposed_queryset': decomposed_queryset,
                'user_id': self.request.user.pk
            }
        )

        messages.success(
            message=_(
                'Event list export task queued successfully.'
            ), request=self.request
        )


class EventListExportView(EventExportBaseView):
    object_permission = permission_events_export
    view_icon = icon_event_list_export

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'title': _('Export events')
            }
        )
        return context

    def get_queryset_parameters(self):
        return {
            '_method_name': 'all'
        }


class ObjectEventExportView(
    ExternalContentTypeObjectViewMixin, EventExportBaseView
):
    view_icon = icon_object_event_list_export

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'object': self.external_object,
                'title': _('Export events of: %s') % self.external_object
            }
        )
        return context

    def get_queryset_parameters(self):
        return {
            '_method_name': 'any', 'obj': self.external_object
        }


class VerbEventExportView(VerbEventViewMixin, EventExportBaseView):
    view_icon = icon_verb_event_list_export

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'title': _(
                    'Export events of type: %s'
                ) % self.event_type
            }
        )
        return context

    def get_queryset_parameters(self):
        return {
            '_method_name': 'filter', 'verb': self.event_type.id
        }



from mayan.apps.testing.tests.base import BaseTestCase

from ..exceptions import DynamicSearchScopedQueryError
from ..literals import (
    ERROR_TEXT_NO_RESULT_SCOPE, SCOPE_DELIMITER, SCOPE_MARKER,
    SCOPE_RESULT_MARKER
)

from .literals import (
    TEST_SCOPED_QUERY_ENTRY_FIELD_NAME,
    TEST_SCOPED_QUERY_ENTRY_FIELD_NAME_INVALID,
    TEST_SCOPED_QUERY_ENTRY_OPERATOR_OPERAND_LIST,
    TEST_SCOPED_QUERY_ENTRY_OPERATOR_OPERATOR_TEXT,
    TEST_SCOPED_QUERY_ENTRY_OPERATOR_SCOPE_IDENTIFIER,
    TEST_SCOPED_QUERY_ENTRY_SCOPE_IDENTIFIER,
    TEST_SCOPED_QUERY_ENTRY_SCOPE_IDENTIFIER_INVALID,
    TEST_SCOPED_QUERY_ENTRY_VALUE, TEST_SCOPED_QUERY_RESULT_SCOPE_IDENTIFIER
)
from .mixins.base import SearchTestMixin, TestSearchObjectHierarchyTestMixin
from .mixins.scoped_query_mixins import ScopedQueryTestMixin


class ScopedQueryEntryFieldTestCase(
    ScopedQueryTestMixin, TestSearchObjectHierarchyTestMixin,
    SearchTestMixin, BaseTestCase
):
    auto_test_search_objects_create = False

    def test_check_filter_scope(self):
        test_scope_entry_count = len(
            self._test_scoped_query.scope_entry_list
        )

        self._add_test_scoped_query_entry_filter()

        self.assertEqual(
            len(self._test_scoped_query.scope_entry_list),
            test_scope_entry_count + 1
        )

        test_scope_entry = self._test_scoped_query.get_scope_entry_by_identifier(
            scope_identifier=self._test_scoped_query_entry.scope_identifier
        )

        self.assertEqual(
            test_scope_entry.field_name, TEST_SCOPED_QUERY_ENTRY_FIELD_NAME
        )
        self.assertEqual(
            test_scope_entry.scope_identifier,
            TEST_SCOPED_QUERY_ENTRY_SCOPE_IDENTIFIER
        )
        self.assertEqual(
            test_scope_entry.to_string(),
            '{scope_marker}{scope_identifier}{scope_delimiter}{field_name}={value}'.format(
                field_name=TEST_SCOPED_QUERY_ENTRY_FIELD_NAME,
                scope_delimiter=SCOPE_DELIMITER,
                scope_identifier=TEST_SCOPED_QUERY_ENTRY_SCOPE_IDENTIFIER,
                scope_marker=SCOPE_MARKER,
                value=TEST_SCOPED_QUERY_ENTRY_VALUE
            )
        )

    def test_check_filter_scope_duplicate_identifier(self):
        test_scope_entry_count = len(self._test_scoped_query.scope_entry_list)

        self._add_test_scoped_query_entry_filter()

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            self._add_test_scoped_query_entry_filter()

        self.assertEqual(
            len(self._test_scoped_query.scope_entry_list),
            test_scope_entry_count + 1
        )

        test_scope_entry = self._test_scoped_query.get_scope_entry_by_identifier(
            scope_identifier=self._test_scoped_query_entry.scope_identifier
        )

        self.assertEqual(
            test_scope_entry.field_name, TEST_SCOPED_QUERY_ENTRY_FIELD_NAME
        )
        self.assertEqual(
            test_scope_entry.to_string(),
            '{scope_marker}{scope_identifier}{scope_delimiter}{field_name}={value}'.format(
                field_name=TEST_SCOPED_QUERY_ENTRY_FIELD_NAME,
                scope_delimiter=SCOPE_DELIMITER,
                scope_identifier=TEST_SCOPED_QUERY_ENTRY_SCOPE_IDENTIFIER,
                scope_marker=SCOPE_MARKER,
                value=TEST_SCOPED_QUERY_ENTRY_VALUE
            )
        )

    def test_check_filter_scope_invalid_field_name(self):
        test_scope_entry_count = len(self._test_scoped_query.scope_entry_list)

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            self._add_test_scoped_query_entry_filter(
                field_name=TEST_SCOPED_QUERY_ENTRY_FIELD_NAME_INVALID
            )

        self.assertEqual(
            len(self._test_scoped_query.scope_entry_list),
            test_scope_entry_count
        )


class ScopedQueryEntryOperatorTestCase(
    ScopedQueryTestMixin, TestSearchObjectHierarchyTestMixin,
    SearchTestMixin, BaseTestCase
):
    auto_test_search_objects_create = False

    def test_check_operator_scope(self):
        test_scope_entry_count = len(
            self._test_scoped_query.scope_entry_list
        )

        self._add_test_scoped_query_entry_operator()

        self.assertEqual(
            len(self._test_scoped_query.scope_entry_list),
            test_scope_entry_count + 1
        )

        test_scope_entry = self._test_scoped_query.get_scope_entry_by_identifier(
            scope_identifier=self._test_scoped_query_entry.scope_identifier
        )

        self.assertEqual(
            test_scope_entry.scope_identifier,
            TEST_SCOPED_QUERY_ENTRY_OPERATOR_SCOPE_IDENTIFIER,
        )
        self.assertEqual(
            tuple(test_scope_entry.operand_list),
            tuple(TEST_SCOPED_QUERY_ENTRY_OPERATOR_OPERAND_LIST)
        )
        self.assertEqual(
            test_scope_entry.operator_text,
            TEST_SCOPED_QUERY_ENTRY_OPERATOR_OPERATOR_TEXT
        )
        self.assertEqual(
            test_scope_entry.to_string(),
            '{scope_marker}{scope_identifier}={scope_operator}{scope_delimiter}{scope_operand_0}{scope_delimiter}{scope_operand_1}'.format(
                scope_delimiter=SCOPE_DELIMITER,
                scope_identifier=TEST_SCOPED_QUERY_ENTRY_OPERATOR_SCOPE_IDENTIFIER,
                scope_marker=SCOPE_MARKER,
                scope_operand_0=TEST_SCOPED_QUERY_ENTRY_OPERATOR_OPERAND_LIST[0],
                scope_operand_1=TEST_SCOPED_QUERY_ENTRY_OPERATOR_OPERAND_LIST[1],
                scope_operator=TEST_SCOPED_QUERY_ENTRY_OPERATOR_OPERATOR_TEXT
            )
        )

    def test_check_operator_scope_self_reference(self):
        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            self._add_test_scoped_query_entry_operator(
                scope_operand_list=(
                    TEST_SCOPED_QUERY_ENTRY_OPERATOR_SCOPE_IDENTIFIER, '99'
                )
            )

    def test_check_operator_scope_single_operand(self):
        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            self._add_test_scoped_query_entry_operator(scope_operand_list=('0',))

    def test_check_operator_scope_three_operands(self):
        test_scope_entry_count = len(
            self._test_scoped_query.scope_entry_list
        )

        self._add_test_scoped_query_entry_operator(
            scope_operand_list=('0', '1', '2')
        )

        self.assertEqual(
            len(self._test_scoped_query.scope_entry_list),
            test_scope_entry_count + 1
        )

        test_scope_entry = self._test_scoped_query.get_scope_entry_by_identifier(
            scope_identifier=self._test_scoped_query_entry.scope_identifier
        )

        self.assertEqual(
            test_scope_entry.scope_identifier,
            TEST_SCOPED_QUERY_ENTRY_OPERATOR_SCOPE_IDENTIFIER
        )
        self.assertEqual(
            tuple(test_scope_entry.operand_list), ('0', '1', '2')
        )
        self.assertEqual(
            test_scope_entry.operator_text,
            TEST_SCOPED_QUERY_ENTRY_OPERATOR_OPERATOR_TEXT
        )
        self.assertEqual(
            test_scope_entry.to_string(),
            '{scope_marker}{scope_identifier}={scope_operator}{scope_delimiter}{scope_operand_0}{scope_delimiter}{scope_operand_1}{scope_delimiter}{scope_operand_2}'.format(
                scope_delimiter=SCOPE_DELIMITER,
                scope_identifier=TEST_SCOPED_QUERY_ENTRY_OPERATOR_SCOPE_IDENTIFIER,
                scope_marker=SCOPE_MARKER,
                scope_operand_0='0',
                scope_operand_1='1',
                scope_operand_2='2',
                scope_operator=TEST_SCOPED_QUERY_ENTRY_OPERATOR_OPERATOR_TEXT
            )
        )


class ScopedQueryEntryResultTestCase(
    ScopedQueryTestMixin, TestSearchObjectHierarchyTestMixin,
    SearchTestMixin, BaseTestCase
):
    auto_test_search_objects_create = False

    def test_check_result_scope(self):
        test_scope_entry_count = len(
            self._test_scoped_query.scope_entry_list
        )

        self._add_test_scoped_query_result_scope(
            value=TEST_SCOPED_QUERY_RESULT_SCOPE_IDENTIFIER
        )

        self.assertEqual(
            len(self._test_scoped_query.scope_entry_list),
            test_scope_entry_count + 1
        )

        self.assertNotEqual(
            self._test_scoped_query.entry_point, None
        )
        self.assertEqual(
            self._test_scoped_query.entry_point.result_scope_identifier,
            TEST_SCOPED_QUERY_RESULT_SCOPE_IDENTIFIER
        )
        self.assertEqual(
            self._test_scoped_query_entry.to_string(),
            '{scope_marker}{scope_result_marker}={result_scope_identifier}'.format(
                result_scope_identifier=TEST_SCOPED_QUERY_RESULT_SCOPE_IDENTIFIER,
                scope_marker=SCOPE_MARKER,
                scope_result_marker=SCOPE_RESULT_MARKER
            )
        )

    def test_no_result_scope(self):
        self._add_test_scoped_query_entry_filter()

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError) as exception_context:
            self._test_scoped_query.do_resolve(
                search_backend=self._test_search_backend
            )
        self.assertEqual(
            str(exception_context.exception), ERROR_TEXT_NO_RESULT_SCOPE
        )

    def test_blank_result_scope(self):
        self._add_test_scoped_query_result_scope(value='')

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError) as exception_context:
            self._test_scoped_query.do_resolve(
                search_backend=self._test_search_backend
            )
        self.assertEqual(
            str(exception_context.exception), ERROR_TEXT_NO_RESULT_SCOPE
        )


class ScopedQueryTestCase(
    ScopedQueryTestMixin, TestSearchObjectHierarchyTestMixin,
    SearchTestMixin, BaseTestCase
):
    auto_test_search_objects_create = False

    def test_scope_query_empty(self):
        self.assertTrue(self._test_scoped_query.is_empty)

    def test_empty_entries(self):
        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            self._add_test_scoped_query_entry_filter(value='')

    def test_empty_quoted_entries(self):
        self._add_test_scoped_query_entry_filter(value='""')
        self.assertFalse(self._test_scoped_query.is_empty)

    def test_get_invalid_entry(self):
        self._add_test_scoped_query_entry_filter()

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            self._test_scoped_query.get_scope_entry_by_identifier(
                scope_identifier=TEST_SCOPED_QUERY_ENTRY_SCOPE_IDENTIFIER_INVALID
            )

    def test_solve_invalid_scope_identifier(self):
        self._add_test_scoped_query_entry_filter()

        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError) as exception_context:
            self._test_scoped_query.do_scope_solve(
                search_backend=self._test_search_backend,
                scope_identifier=TEST_SCOPED_QUERY_ENTRY_SCOPE_IDENTIFIER_INVALID
            )

        self.assertTrue(
            'No scope found with identifier' in str(exception_context.exception)
        )


class ScopedQueryDataFilterEntryQueryTypeTestCase(
    ScopedQueryTestMixin, TestSearchObjectHierarchyTestMixin,
    SearchTestMixin, BaseTestCase
):
    auto_test_search_objects_create = False

    def test_scoped_query_data_filter_double_words(self):
        with self.assertRaises(expected_exception=DynamicSearchScopedQueryError):
            self._add_test_scoped_query_entry_filter(value='2021 2022')

    def test_scoped_query_data_filter_double_words_quoted(self):
        test_scope_entry_count = len(
            self._test_scoped_query.scope_entry_list
        )

        self._add_test_scoped_query_entry_filter(value='"2021 2022"')

        self.assertEqual(
            len(self._test_scoped_query.scope_entry_list),
            test_scope_entry_count + 1
        )

        test_scope_entry = self._test_scoped_query.get_scope_entry_by_identifier(
            scope_identifier=self._test_scoped_query_entry.scope_identifier
        )

        self.assertEqual(test_scope_entry.is_quoted_value, True)
        self.assertEqual(test_scope_entry.is_raw_value, False)
        self.assertEqual(test_scope_entry.value, '2021 2022')

    def test_scoped_query_data_filter_raw_value_single_word(self):
        test_scope_entry_count = len(
            self._test_scoped_query.scope_entry_list
        )

        self._add_test_scoped_query_entry_filter(value='`2021`')

        self.assertEqual(
            len(self._test_scoped_query.scope_entry_list),
            test_scope_entry_count + 1
        )

        test_scope_entry = self._test_scoped_query.get_scope_entry_by_identifier(
            scope_identifier=self._test_scoped_query_entry.scope_identifier
        )

        self.assertEqual(test_scope_entry.is_quoted_value, False)
        self.assertEqual(test_scope_entry.is_raw_value, True)
        self.assertEqual(test_scope_entry.value, '2021')

    def test_scoped_query_data_filter_raw_value_double_words(self):
        test_scope_entry_count = len(
            self._test_scoped_query.scope_entry_list
        )

        self._add_test_scoped_query_entry_filter(value='`2021 2022`')

        self.assertEqual(
            len(self._test_scoped_query.scope_entry_list),
            test_scope_entry_count + 1
        )

        test_scope_entry = self._test_scoped_query.get_scope_entry_by_identifier(
            scope_identifier=self._test_scoped_query_entry.scope_identifier
        )

        self.assertEqual(test_scope_entry.is_quoted_value, False)
        self.assertEqual(test_scope_entry.is_raw_value, True)
        self.assertEqual(test_scope_entry.value, '2021 2022')



from mayan.apps.common.tests.literals import TEST_BINARY_CONTENT
from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import (
    event_download_file_deleted, event_download_file_downloaded
)
from ..models import DownloadFile
from ..permissions import (
    permission_download_file_delete, permission_download_file_download,
    permission_download_file_view
)

from .mixins import DownloadFileTestMixin, DownloadFileViewTestMixin


class DownloadFileViewTestCase(
    DownloadFileTestMixin, DownloadFileViewTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_user()

    def test_download_file_delete_view_not_owner_no_permission(self):
        self._create_test_download_file(user=self._test_user)

        download_file_count = DownloadFile.objects.count()

        self._clear_events()

        response = self._request_test_download_file_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_download_file_delete_view_not_owner_with_access(self):
        self._create_test_download_file(user=self._test_user)

        self.grant_access(
            obj=self._test_download_file,
            permission=permission_download_file_delete
        )

        download_file_count = DownloadFile.objects.count()

        self._clear_events()

        response = self._request_test_download_file_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, None)
        self.assertEqual(events[0].verb, event_download_file_deleted.id)

    def test_download_file_delete_view_owner_no_permission(self):
        self._create_test_download_file()

        download_file_count = DownloadFile.objects.count()

        self._clear_events()

        response = self._request_test_download_file_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, None)
        self.assertEqual(events[0].verb, event_download_file_deleted.id)

    def test_download_file_delete_view_owner_with_access(self):
        self._create_test_download_file()

        self.grant_access(
            obj=self._test_download_file,
            permission=permission_download_file_delete
        )

        download_file_count = DownloadFile.objects.count()

        self._clear_events()

        response = self._request_test_download_file_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, None)
        self.assertEqual(events[0].verb, event_download_file_deleted.id)

    def test_download_file_download_view_not_owner_no_permission(self):
        self._create_test_download_file(
            content=TEST_BINARY_CONTENT, user=self._test_user
        )

        self._clear_events()

        response = self._request_test_download_file_download_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_download_file_download_view_not_owner_with_access(self):
        # Set the expected_content_types for
        # common.tests.mixins.ContentTypeCheckMixin
        self.expected_content_types = ('text/plain',)

        self._create_test_download_file(
            content=TEST_BINARY_CONTENT, user=self._test_user
        )

        self.grant_access(
            obj=self._test_download_file,
            permission=permission_download_file_download
        )

        self._clear_events()

        response = self._request_test_download_file_download_view()
        self.assertEqual(response.status_code, 200)

        with self._test_download_file.open(mode='rb') as file_object:
            self.assert_download_response(
                response=response, content=file_object.read(),
                filename=str(self._test_download_file),
                mime_type='text/plain'
            )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_download_file)
        self.assertEqual(events[0].verb, event_download_file_downloaded.id)

    def test_download_file_download_view_owner_no_permission(self):
        # Set the expected_content_types for
        # common.tests.mixins.ContentTypeCheckMixin
        self.expected_content_types = ('text/plain',)

        self._create_test_download_file(content=TEST_BINARY_CONTENT)

        self._clear_events()

        response = self._request_test_download_file_download_view()
        self.assertEqual(response.status_code, 200)

        with self._test_download_file.open(mode='rb') as file_object:
            self.assert_download_response(
                response=response, content=file_object.read(),
                filename=str(self._test_download_file),
                mime_type='text/plain'
            )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_download_file)
        self.assertEqual(events[0].verb, event_download_file_downloaded.id)

    def test_download_file_download_view_owner_with_access(self):
        # Set the expected_content_types for
        # common.tests.mixins.ContentTypeCheckMixin
        self.expected_content_types = ('text/plain',)

        self._create_test_download_file(content=TEST_BINARY_CONTENT)

        self.grant_access(
            obj=self._test_download_file,
            permission=permission_download_file_download
        )

        self._clear_events()

        response = self._request_test_download_file_download_view()
        self.assertEqual(response.status_code, 200)

        with self._test_download_file.open(mode='rb') as file_object:
            self.assert_download_response(
                response=response, content=file_object.read(),
                filename=str(self._test_download_file),
                mime_type='text/plain'
            )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_download_file)
        self.assertEqual(events[0].verb, event_download_file_downloaded.id)

    def test_download_file_list_view_not_owner_no_permission(self):
        self._create_test_download_file(user=self._test_user)

        self._clear_events()

        response = self._request_test_download_file_list_view()
        self.assertNotContains(
            response=response, text=str(self._test_download_file),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_download_file_list_view_not_owner_with_access(self):
        self._create_test_download_file(user=self._test_user)

        self.grant_access(
            obj=self._test_download_file,
            permission=permission_download_file_view
        )

        self._clear_events()

        response = self._request_test_download_file_list_view()

        self.assertContains(
            response=response, text=str(self._test_download_file),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_download_file_list_view_owner_no_permission(self):
        self._create_test_download_file()

        self._clear_events()

        response = self._request_test_download_file_list_view()
        self.assertContains(
            response=response, text=str(self._test_download_file),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_download_file_list_view_owner_with_access(self):
        self._create_test_download_file()

        self.grant_access(
            obj=self._test_download_file,
            permission=permission_download_file_view
        )

        self._clear_events()

        response = self._request_test_download_file_list_view()
        self.assertContains(
            response=response, text=str(self._test_download_file),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)
